#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <richedit.h>
#include "AkelDLL.h"
#include "Resources\Resource.h"


//Defines
#define STRID_SETUP      1
#define STRID_SEPARATOR  2
#define STRID_LEGEND     3
#define STRID_OK         4
#define STRID_CANCEL     5

#define BUFFER_SIZE      1024

#define AKDLL_SETUP   (WM_USER + 1)

#define OF_WINDOWRECT     0x1
#define OF_SETTINGS       0x2

//DestroyDock type
#define DKT_UNLOAD         0
#define DKT_NOUNLOAD       1
#define DKT_ONMAINFINISH   2

//DIALOGRESIZE
#define DRS_SIZE  0x1 //Resize control
#define DRS_MOVE  0x2 //Move control
#define DRS_X     0x4 //X value
#define DRS_Y     0x8 //Y value

typedef struct {
  HWND *lpWnd;   //Control window
  DWORD dwType;  //See DRS_* defines
} DIALOGRESIZE;

//Functions prototypes
void CreateDock(HWND *hWndDock, DOCK **dkDock);
void DestroyDock(HWND hWndDock, DWORD dwType);
BOOL CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc);
BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL EscapeStringToEscapeDataA(char *pInput, char *szOutput);
BOOL EscapeStringToEscapeDataW(wchar_t *wpInput, wchar_t *wszOutput);
void EscapeDataToEscapeStringA(char *pInput, char *szOutput);
void EscapeDataToEscapeStringW(wchar_t *wpInput, wchar_t *wszOutput);
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA(DWORD dwFlags);
void SaveOptionsW(DWORD dwFlags);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceDLL;
HWND hMainWnd;
HWND hWndEdit;
HMENU hPopupEdit;
HICON hMainIcon;
BOOL bOldWindows;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
char szSeparator[MAX_PATH]="\r------------------------------------------------------------\r";
wchar_t wszSeparator[MAX_PATH]=L"\r------------------------------------------------------------\r";
char szBuffer[BUFFER_SIZE];
wchar_t wszBuffer[BUFFER_SIZE];
HWND hWndNextViewer=NULL;
HWND hWndMainDlg=NULL;
HWND hWndCapture=NULL;
RECT rcMainInitDialog={0};
RECT rcMainCurrentDialog={0};
RECT rcDockRect={0};
int nDockSide=DKS_BOTTOM;
DOCK *dkMainDlg=NULL;
BOOL bDockWaitResize=FALSE;
BOOL bMainDialogRectSave=FALSE;
WNDPROCDATA *NewMainProcData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="CaptureClipboard";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  //Is plugin already loaded?
  if (bInitMain)
  {
    if (hWndMainDlg) DestroyDock(hWndMainDlg, DKT_UNLOAD);

    //Stay in memory, but show as non-active
    pd->nUnload=UD_NONUNLOAD_NONACTIVE;
  }
  else
  {
    InitMain();

    bDockWaitResize=pd->bOnStart;
    CreateDock(&hWndMainDlg, &dkMainDlg);

    //Stay in memory, and show as active
    pd->nUnload=UD_NONUNLOAD_ACTIVE;
  }
}

void CreateDock(HWND *hWndDock, DOCK **dkDock)
{
  DOCK dk={0};

  *hWndDock=NULL;
  dk.dwFlags=DKF_DRAGDROP;
  dk.nSide=nDockSide;
  dk.rcSize=rcDockRect;

  if (*dkDock=(DOCK *)SendMessage(hMainWnd, AKD_DOCK, DK_ADD, (LPARAM)&dk))
  {
    if (bOldWindows)
      *hWndDock=CreateDialogA(hInstanceDLL, MAKEINTRESOURCEA(IDD_CAPTURE), hMainWnd, (DLGPROC)MainDlgProcA);
    else
      *hWndDock=CreateDialogW(hInstanceDLL, MAKEINTRESOURCEW(IDD_CAPTURE), hMainWnd, (DLGPROC)MainDlgProcW);

    if (*hWndDock)
    {
      if (!(*dkDock)->rcSize.right || !(*dkDock)->rcSize.bottom)
      {
        GetWindowRect(*hWndDock, &(*dkDock)->rcSize);
        (*dkDock)->rcSize.right-=(*dkDock)->rcSize.left;
        (*dkDock)->rcSize.bottom-=(*dkDock)->rcSize.top;
        (*dkDock)->rcSize.left=0;
        (*dkDock)->rcSize.top=0;
      }
      (*dkDock)->hWnd=*hWndDock;
      SendMessage(hMainWnd, AKD_DOCK, DK_SUBCLASS, (LPARAM)*dkDock);

      if (!bDockWaitResize)
      {
        SendMessage(hMainWnd, AKD_RESIZE, 0, 0);
        ShowWindow(*hWndDock, SW_SHOW);
      }
    }
  }
}

void DestroyDock(HWND hWndDock, DWORD dwType)
{
  SendMessage(hWndDock, WM_COMMAND, IDC_CLOSE, dwType);
}

BOOL CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndTitleText;
  static HWND hWndTitleClose;
  static HMENU hMenuLabel;
  static DIALOGRESIZE drs[]={{&hWndTitleText,    DRS_SIZE|DRS_X},
                             {&hWndTitleClose,   DRS_MOVE|DRS_X},
                             {&hWndCapture,      DRS_SIZE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    hWndCapture=GetDlgItem(hDlg, IDC_CAPTURE);
    hWndTitleText=GetDlgItem(hDlg, IDC_TITLETEXT);
    hWndTitleClose=GetDlgItem(hDlg, IDC_CLOSE);

    //SendMessage(hWndCapture, EM_SETUNDOLIMIT, 0, 0);

    hMenuLabel=CreatePopupMenu();
    AppendMenuA(hMenuLabel, MF_STRING, IDM_SETUP, GetLangStringA(wLangSystem, STRID_SETUP));

    SetWindowTextA(hWndTitleText, szPluginName);
    SendMessage(hMainWnd, AKD_SETCLOSEBUTTON, (WPARAM)hWndTitleClose, (LPARAM)NULL);
  }
  else if (uMsg == AKDLL_SETUP)
  {
    DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SETUP), hDlg, (DLGPROC)SetupDlgProcA);
  }
  else if (uMsg == WM_LBUTTONDBLCLK)
  {
    POINT pt;
    RECT rc;

    GetCursorPos(&pt);
    GetWindowRect(hWndTitleText, &rc);

    if (PtInRect(&rc, pt))
    {
      PostMessage(hDlg, AKDLL_SETUP, 0, 0);
    }
  }
  else if (uMsg == WM_CONTEXTMENU)
  {
    if ((HWND)wParam == hDlg)
    {
      POINT pt;
      RECT rc;

      GetCursorPos(&pt);
      GetWindowRect(hWndTitleText, &rc);

      if (PtInRect(&rc, pt))
      {
        int nCmd;

        nCmd=TrackPopupMenu(hMenuLabel, TPM_RETURNCMD|TPM_NONOTIFY|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hWndTitleText, NULL);

        if (nCmd == IDM_SETUP)
          PostMessage(hDlg, AKDLL_SETUP, 0, 0);
      }
    }
    else if ((HWND)wParam == hWndCapture)
    {
      RECT rcRect;
      POINT pt;
      CHARRANGE cr;
      int nCmd;

      if (lParam == -1)
      {
        GetWindowRect((HWND)wParam, &rcRect);
        pt.x=rcRect.left;
        pt.y=rcRect.top;
      }
      else
      {
        GetCursorPos(&pt);
      }
      SendMessage((HWND)wParam, EM_EXGETSEL, 0, (LPARAM)&cr);
      SendMessage((HWND)wParam, EM_EXGETSEL, 0, (LPARAM)&cr);
      EnableMenuItem(hPopupEdit, IDM_EDIT_UNDO, SendMessage((HWND)wParam, EM_CANUNDO, 0, 0)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_REDO, SendMessage((HWND)wParam, EM_CANREDO, 0, 0)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_CUT, (cr.cpMin < cr.cpMax)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_COPY, (cr.cpMin < cr.cpMax)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_PASTE, SendMessage((HWND)wParam, EM_CANPASTE, 0, 0)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_CLEAR, (cr.cpMin < cr.cpMax)?MF_ENABLED:MF_GRAYED);

      nCmd=TrackPopupMenu(hPopupEdit, TPM_RETURNCMD|TPM_NONOTIFY|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hDlg, NULL);

      if (nCmd == IDM_EDIT_UNDO)
        SendMessage((HWND)wParam, WM_UNDO, 0, 0);
      else if (nCmd == IDM_EDIT_REDO)
        SendMessage((HWND)wParam, EM_REDO, 0, 0);
      else if (nCmd == IDM_EDIT_CUT)
        SendMessage((HWND)wParam, WM_CUT, 0, 0);
      else if (nCmd == IDM_EDIT_COPY)
        SendMessage((HWND)wParam, WM_COPY, 0, 0);
      else if (nCmd == IDM_EDIT_PASTE)
        SendMessage((HWND)wParam, WM_PASTE, 0, 0);
      else if (nCmd == IDM_EDIT_CLEAR)
        SendMessage((HWND)wParam, WM_CLEAR, 0, 0);
      else if (nCmd == IDM_EDIT_SELECTALL)
        SendMessage((HWND)wParam, EM_SETSEL, 0, -1);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_CLOSE ||
        LOWORD(wParam) == IDCANCEL)
    {
      if (bMainDialogRectSave || dkMainDlg->nSide != nDockSide)
        SaveOptionsA(OF_WINDOWRECT);

      if (lParam != DKT_NOUNLOAD)
      {
        UninitMain();
      }
      SendMessage(hMainWnd, AKD_DOCK, DK_DELETE, (LPARAM)dkMainDlg);
      dkMainDlg=NULL;

      DestroyMenu(hMenuLabel);
      DestroyWindow(hWndMainDlg);
      hWndMainDlg=NULL;

      if (lParam != DKT_ONMAINFINISH)
      {
        SendMessage(hMainWnd, AKD_RESIZE, 0, 0);
        SetFocus(hMainWnd);
        if (lParam != DKT_NOUNLOAD)
          PostMessage(hMainWnd, AKD_DLLUNLOAD, (WPARAM)hInstanceDLL, (LPARAM)NULL);
      }
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
  {
    if (IsWindowVisible(hDlg)) bMainDialogRectSave=TRUE;
    if (dkMainDlg) GetWindowPos(hWndTitleText, hDlg, &dkMainDlg->rcDragDrop);
  }

  return FALSE;
}

BOOL CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndTitleText;
  static HWND hWndTitleClose;
  static HMENU hMenuLabel;
  static DIALOGRESIZE drs[]={{&hWndTitleText,    DRS_SIZE|DRS_X},
                             {&hWndTitleClose,   DRS_MOVE|DRS_X},
                             {&hWndCapture,      DRS_SIZE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    hWndCapture=GetDlgItem(hDlg, IDC_CAPTURE);
    hWndTitleText=GetDlgItem(hDlg, IDC_TITLETEXT);
    hWndTitleClose=GetDlgItem(hDlg, IDC_CLOSE);

    //SendMessage(hWndCapture, EM_SETUNDOLIMIT, 0, 0);

    hMenuLabel=CreatePopupMenu();
    AppendMenuW(hMenuLabel, MF_STRING, IDM_SETUP, GetLangStringW(wLangSystem, STRID_SETUP));

    SetWindowTextW(hWndTitleText, wszPluginName);
    SendMessage(hMainWnd, AKD_SETCLOSEBUTTON, (WPARAM)hWndTitleClose, (LPARAM)NULL);
  }
  else if (uMsg == AKDLL_SETUP)
  {
    DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SETUP), hDlg, (DLGPROC)SetupDlgProcW);
  }
  else if (uMsg == WM_LBUTTONDBLCLK)
  {
    POINT pt;
    RECT rc;

    GetCursorPos(&pt);
    GetWindowRect(hWndTitleText, &rc);

    if (PtInRect(&rc, pt))
    {
      PostMessage(hDlg, AKDLL_SETUP, 0, 0);
    }
  }
  else if (uMsg == WM_CONTEXTMENU)
  {
    if ((HWND)wParam == hDlg)
    {
      POINT pt;
      RECT rc;

      GetCursorPos(&pt);
      GetWindowRect(hWndTitleText, &rc);

      if (PtInRect(&rc, pt))
      {
        int nCmd;

        nCmd=TrackPopupMenu(hMenuLabel, TPM_RETURNCMD|TPM_NONOTIFY|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hWndTitleText, NULL);

        if (nCmd == IDM_SETUP)
          PostMessage(hDlg, AKDLL_SETUP, 0, 0);
      }
    }
    else if ((HWND)wParam == hWndCapture)
    {
      RECT rcRect;
      POINT pt;
      CHARRANGE cr;
      int nCmd;

      if (lParam == -1)
      {
        GetWindowRect((HWND)wParam, &rcRect);
        pt.x=rcRect.left;
        pt.y=rcRect.top;
      }
      else
      {
        GetCursorPos(&pt);
      }
      SendMessage((HWND)wParam, EM_EXGETSEL, 0, (LPARAM)&cr);
      EnableMenuItem(hPopupEdit, IDM_EDIT_UNDO, SendMessage((HWND)wParam, EM_CANUNDO, 0, 0)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_REDO, SendMessage((HWND)wParam, EM_CANREDO, 0, 0)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_CUT, (cr.cpMin < cr.cpMax)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_COPY, (cr.cpMin < cr.cpMax)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_PASTE, SendMessage((HWND)wParam, EM_CANPASTE, 0, 0)?MF_ENABLED:MF_GRAYED);
      EnableMenuItem(hPopupEdit, IDM_EDIT_CLEAR, (cr.cpMin < cr.cpMax)?MF_ENABLED:MF_GRAYED);

      nCmd=TrackPopupMenu(hPopupEdit, TPM_RETURNCMD|TPM_NONOTIFY|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hDlg, NULL);

      if (nCmd == IDM_EDIT_UNDO)
        SendMessage((HWND)wParam, WM_UNDO, 0, 0);
      else if (nCmd == IDM_EDIT_REDO)
        SendMessage((HWND)wParam, EM_REDO, 0, 0);
      else if (nCmd == IDM_EDIT_CUT)
        SendMessage((HWND)wParam, WM_CUT, 0, 0);
      else if (nCmd == IDM_EDIT_COPY)
        SendMessage((HWND)wParam, WM_COPY, 0, 0);
      else if (nCmd == IDM_EDIT_PASTE)
        SendMessage((HWND)wParam, WM_PASTE, 0, 0);
      else if (nCmd == IDM_EDIT_CLEAR)
        SendMessage((HWND)wParam, WM_CLEAR, 0, 0);
      else if (nCmd == IDM_EDIT_SELECTALL)
        SendMessage((HWND)wParam, EM_SETSEL, 0, -1);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_CLOSE ||
        LOWORD(wParam) == IDCANCEL)
    {
      if (bMainDialogRectSave || dkMainDlg->nSide != nDockSide)
        SaveOptionsW(OF_WINDOWRECT);

      if (lParam != DKT_NOUNLOAD)
      {
        UninitMain();
      }
      SendMessage(hMainWnd, AKD_DOCK, DK_DELETE, (LPARAM)dkMainDlg);
      dkMainDlg=NULL;

      DestroyMenu(hMenuLabel);
      DestroyWindow(hWndMainDlg);
      hWndMainDlg=NULL;

      if (lParam != DKT_ONMAINFINISH)
      {
        SendMessage(hMainWnd, AKD_RESIZE, 0, 0);
        SetFocus(hMainWnd);
        if (lParam != DKT_NOUNLOAD)
          PostMessage(hMainWnd, AKD_DLLUNLOAD, (WPARAM)hInstanceDLL, (LPARAM)NULL);
      }
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
  {
    if (IsWindowVisible(hDlg)) bMainDialogRectSave=TRUE;
    if (dkMainDlg) GetWindowPos(hWndTitleText, hDlg, &dkMainDlg->rcDragDrop);
  }

  return FALSE;
}

BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_INITDIALOG)
  {
    RECT rcTemplate;

    GetWindowPos(hDlg, NULL, rcInit);
    rcTemplate=*rcCurrent;
    *rcCurrent=*rcInit;

    if (rcTemplate.right && rcTemplate.bottom)
    {
      rcTemplate.left=rcInit->left + (rcInit->right - rcTemplate.right) / 2;
      rcTemplate.top=rcInit->top + (rcInit->bottom - rcTemplate.bottom) / 2;
      SetWindowPos(hDlg, 0, rcTemplate.left, rcTemplate.top, rcTemplate.right, rcTemplate.bottom, SWP_NOZORDER);
    }
  }
  else if (uMsg == WM_GETMINMAXINFO)
  {
    //MINMAXINFO *mmi=(MINMAXINFO *)lParam;
    //
    //mmi->ptMinTrackSize.x=rcInit->right;
    //mmi->ptMinTrackSize.y=rcInit->bottom;
  }
  else if (uMsg == WM_SIZE)
  {
    if (lParam)
    {
      RECT rcTemplate;
      RECT rcControl;
      DWORD dwFlags;
      POINT pt;
      int i;

      GetWindowPos(hDlg, NULL, &rcTemplate);
      pt.x=rcTemplate.right - rcCurrent->right;
      pt.y=rcTemplate.bottom - rcCurrent->bottom;
      *rcCurrent=rcTemplate;

      for (i=0; drs[i].lpWnd; ++i)
      {
        if (*drs[i].lpWnd)
        {
          dwFlags=0;
          if (drs[i].dwType & DRS_SIZE)
            dwFlags|=SWP_NOMOVE;
          else if (drs[i].dwType & DRS_MOVE)
            dwFlags|=SWP_NOSIZE;
          GetWindowPos(*drs[i].lpWnd, hDlg, &rcControl);
          SetWindowPos(*drs[i].lpWnd, 0, rcControl.left + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.top + ((drs[i].dwType & DRS_Y)?pt.y:0), rcControl.right + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.bottom + ((drs[i].dwType & DRS_Y)?pt.y:0), dwFlags|SWP_NOZORDER);
        }
      }
      InvalidateRect(hDlg, NULL, TRUE);
      return TRUE;
    }
  }
  else if (uMsg == WM_PAINT)
  {
    //PAINTSTRUCT ps;
    //RECT rcGrip;
    //HDC hDC;
    //
    //if (hDC=BeginPaint(hDlg, &ps))
    //{
    //  GetClientRect(hDlg, &rcGrip);
    //  rcGrip.left=rcGrip.right - GetSystemMetrics(SM_CXVSCROLL);
    //  rcGrip.top=rcGrip.bottom - GetSystemMetrics(SM_CYVSCROLL);
    //  DrawFrameControl(hDC, &rcGrip, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
    //  EndPaint(hDlg, &ps);
    //}
  }
  return FALSE;
}

BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc)
{
  if (GetWindowRect(hWnd, rc))
  {
    rc->right-=rc->left;
    rc->bottom-=rc->top;

    if (hWndOwner)
    {
      if (!ScreenToClient(hWndOwner, (POINT *)&rc->left))
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndSeparator;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndSeparator=GetDlgItem(hDlg, IDC_SETUP_SEPARATOR);

    SetDlgItemTextA(hDlg, IDC_SETUP_SEPARATOR_LABEL, GetLangStringA(wLangSystem, STRID_SEPARATOR));
    SetDlgItemTextA(hDlg, IDC_SETUP_SEPARATOR_LEGEND, GetLangStringA(wLangSystem, STRID_LEGEND));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    SendMessage(hWndSeparator, EM_LIMITTEXT, MAX_PATH, 0);
    EscapeDataToEscapeStringA(szSeparator, szBuffer);
    SetWindowTextA(hWndSeparator, szBuffer);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK)
    {
      GetWindowTextA(hWndSeparator, szBuffer, MAX_PATH);
      EscapeStringToEscapeDataA(szBuffer, szSeparator);

      SaveOptionsA(OF_SETTINGS);

      EndDialog(hDlg, 0);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndSeparator;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndSeparator=GetDlgItem(hDlg, IDC_SETUP_SEPARATOR);

    SetDlgItemTextW(hDlg, IDC_SETUP_SEPARATOR_LABEL, GetLangStringW(wLangSystem, STRID_SEPARATOR));
    SetDlgItemTextW(hDlg, IDC_SETUP_SEPARATOR_LEGEND, GetLangStringW(wLangSystem, STRID_LEGEND));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    SendMessage(hWndSeparator, EM_LIMITTEXT, MAX_PATH, 0);
    EscapeDataToEscapeStringW(wszSeparator, wszBuffer);
    SetWindowTextW(hWndSeparator, wszBuffer);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK)
    {
      GetWindowTextW(hWndSeparator, wszBuffer, MAX_PATH);
      EscapeStringToEscapeDataW(wszBuffer, wszSeparator);

      SaveOptionsW(OF_SETTINGS);

      EndDialog(hDlg, 0);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_CHANGECBCHAIN)
  {
    if (hWndNextViewer == (HWND)wParam)
      hWndNextViewer=(HWND)lParam;
    else if (hWndNextViewer && IsWindow(hWndNextViewer))
      SendMessage(hWndNextViewer, uMsg, wParam, lParam);
  }
  else if (uMsg == WM_DRAWCLIPBOARD)
  {
    if (hWndMainDlg)
    {
      if (GetFocus() != hWndCapture)
      {
        //With AkelEdit 2.0 and higher this line non needed
        //SendMessage(hWndCapture, EM_SETREADONLY, (WPARAM)FALSE, 0);

        SendMessage(hWndCapture, EM_SETSEL, (WPARAM)0x7FFFFFFF, (LPARAM)0x7FFFFFFF);
        SendMessage(hMainWnd, AKD_PASTE, (WPARAM)hWndCapture, FALSE);

        if (bOldWindows)
          SendMessageA(hWndCapture, EM_REPLACESEL, (WPARAM)TRUE, (LPARAM)szSeparator);
        else
          SendMessageW(hWndCapture, EM_REPLACESEL, (WPARAM)TRUE, (LPARAM)wszSeparator);

        //With AkelEdit 2.0 and higher this line non needed
        //SendMessage(hWndCapture, EM_SETREADONLY, (WPARAM)TRUE, 0);
      }
    }
    if (hWndNextViewer && IsWindow(hWndNextViewer))
      SendMessage(hWndNextViewer, uMsg, wParam, lParam);
  }
  else if (uMsg == AKDN_SIZE)
  {
    if (bDockWaitResize)
    {
      bDockWaitResize=FALSE;
      ShowWindow(hWndMainDlg, SW_SHOW);
    }
  }
  else if (uMsg == AKDN_MAIN_ONFINISH)
  {
    NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);

    if (hWndMainDlg)
    {
      DestroyDock(hWndMainDlg, DKT_ONMAINFINISH);
    }
    return FALSE;
  }

  //Call next procedure
  return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

BOOL EscapeStringToEscapeDataA(char *pInput, char *szOutput)
{
  char *a=pInput;
  char *b=szOutput;

  for (; *a; ++a, ++b)
  {
    if (*a == '\\')
    {
      if (*++a == '\\') *b='\\';
      else if (*a == 'r') *b='\r';
      else if (*a == 'n') *b='\n';
      else if (*a == 't') *b='\t';
      else
      {
        *b='\0';
        return FALSE;
      }
    }
    else *b=*a;
  }
  *b='\0';
  return TRUE;
}

BOOL EscapeStringToEscapeDataW(wchar_t *wpInput, wchar_t *wszOutput)
{
  wchar_t *a=wpInput;
  wchar_t *b=wszOutput;

  for (; *a; ++a, ++b)
  {
    if (*a == '\\')
    {
      if (*++a == '\\') *b='\\';
      else if (*a == 'r') *b='\r';
      else if (*a == 'n') *b='\n';
      else if (*a == 't') *b='\t';
      else
      {
        *b='\0';
        return FALSE;
      }
    }
    else *b=*a;
  }
  *b='\0';
  return TRUE;
}

void EscapeDataToEscapeStringA(char *pInput, char *szOutput)
{
  char *a=pInput;
  char *b=szOutput;

  for (; *a; ++a, ++b)
  {
    if (*a == '\t') *b='\\', *++b='t';
    else if (*a == '\r') *b='\\', *++b='r';
    else if (*a == '\n') *b='\\', *++b='n';
    else if (*a == '\\') *b='\\', *++b='\\';
    else *b=*a;
  }
  *b='\0';
}

void EscapeDataToEscapeStringW(wchar_t *wpInput, wchar_t *wszOutput)
{
  wchar_t *a=wpInput;
  wchar_t *b=wszOutput;

  for (; *a; ++a, ++b)
  {
    if (*a == '\t') *b='\\', *++b='t';
    else if (*a == '\r') *b='\\', *++b='r';
    else if (*a == '\n') *b='\\', *++b='n';
    else if (*a == '\\') *b='\\', *++b='\\';
    else *b=*a;
  }
  *b='\0';
}


//// Options

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="DockSide";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nDockSide;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="DockRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcDockRect;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="Separator";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)szSeparator;
    po.dwData=MAX_PATH;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"DockSide";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nDockSide;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"DockRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcDockRect;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"Separator";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)wszSeparator;
    po.dwData=MAX_PATH * sizeof(wchar_t);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA(DWORD dwFlags)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName="DockSide";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dkMainDlg->nSide;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="DockRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&dkMainDlg->rcSize;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_SETTINGS)
    {
      po.pOptionName="Separator";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)szSeparator;
      po.dwData=lstrlenA(szSeparator) + 1;
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW(DWORD dwFlags)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName=L"DockSide";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dkMainDlg->nSide;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"DockRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&dkMainDlg->rcSize;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_SETTINGS)
    {
      po.pOptionName=L"Separator";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)wszSeparator;
      po.dwData=lstrlenW(wszSeparator) * sizeof(wchar_t) + 2;
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_SETUP)
      return L"\x041D\x0430\x0441\x0442\x0440\x043E\x0439\x043A\x0438\x002E\x002E\x002E";
    if (nStringID == STRID_SEPARATOR)
      return L"\x0420\x0430\x0437\x0434\x0435\x043B\x0438\x0442\x0435\x043B\x044C\x003A";
    if (nStringID == STRID_LEGEND)
      return L" \\\\ - \x043E\x0431\x0440\x0430\x0442\x043D\x044B\x0439\x0020\x0441\x043B\x044D\x0448\n \\r - \x043A\x043E\x043D\x0435\x0446\x0020\x0441\x0442\x0440\x043E\x043A\x0438\n \\t - \x0437\x043D\x0430\x043A\x0020\x0442\x0430\x0431\x0443\x043B\x044F\x0446\x0438\x0438";
    if (nStringID == STRID_OK)
      return L"\x004F\x004B";
    if (nStringID == STRID_CANCEL)
      return L"\x041E\x0442\x043C\x0435\x043D\x0430";
  }
  else
  {
    if (nStringID == STRID_SETUP)
      return L"Settings...";
    if (nStringID == STRID_SEPARATOR)
      return L"Separator:";
    if (nStringID == STRID_LEGEND)
      return L" \\\\ - backslash\n \\r - line feed\n \\t - tabulation";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
  }
  return L"";
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hMainWnd=pd->hMainWnd;
  hWndEdit=pd->hWndEdit;
  hMainIcon=pd->hMainIcon;
  bOldWindows=pd->bOldWindows;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
    ReadOptionsA();
  else
    ReadOptionsW();

  //Context menu
  hPopupEdit=GetSubMenu(pd->hPopupMenu, 0);
}

void InitMain()
{
  bInitMain=TRUE;

  //View clipboard
  hWndNextViewer=SetClipboardViewer(hMainWnd);

  //SubClass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewMainProc, (LPARAM)&NewMainProcData);
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }

  //Stop view clipboard
  ChangeClipboardChain(hMainWnd, hWndNextViewer);
  hWndNextViewer=NULL;
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
  }
  return TRUE;
}
